@@ -274,7 +274,7 @@ do_output (PangoContext     *context,
 
   set_transform (context, transform_cb, cb_context, cb_data, NULL);
 
-  pango_context_set_language (context, pango_language_from_string ("en_US"));
+  pango_context_set_language (context, pango_language_get_default ());
   pango_context_set_base_dir (context,
 			      opt_rtl ? PANGO_DIRECTION_RTL : PANGO_DIRECTION_LTR);
 
@@ -560,6 +560,10 @@ show_version(const char *name,
 {
   g_printf("%s (%s) %s\n", prog_name, PACKAGE_NAME, PACKAGE_VERSION);
   g_printf("module interface version: %s\n", MODULE_VERSION);
+
+  if (PANGO_VERSION != pango_version())
+    g_printf("\nLinked Pango library has a different version: %s\n", pango_version_string ());
+
   exit(0);
 }
 
