@@ -21,9 +21,9 @@
 
 #include <config.h>
 #include <glib.h>
-#include <pango/pango-glyph.h>
-#include <pango/pango-font.h>
-#include <pango/pango-utils.h>
+#include "pango-glyph.h"
+#include "pango-font.h"
+#include "pango-impl-utils.h"
 
 /**
  * pango_glyph_string_new:
@@ -67,7 +67,7 @@ pango_glyph_string_set_size (PangoGlyphString *string, gint new_len)
       
       if (string->space < 0)
         {
-	  g_critical ("glyph string length overflows maximum integer size, truncated");
+	  g_warning ("glyph string length overflows maximum integer size, truncated");
 	  new_len = string->space = G_MAXINT - 8;
 	}
     }
@@ -166,6 +166,9 @@ pango_glyph_string_extents_range (PangoGlyphString *glyphs,
    */
   g_return_if_fail (start <= end);
 
+  if (G_UNLIKELY (!ink_rect && !logical_rect))
+    return;
+
   if (ink_rect)
     {
       ink_rect->x = 0;
@@ -197,9 +200,9 @@ pango_glyph_string_extents_range (PangoGlyphString *glyphs,
 	{
 	  if (ink_rect->width == 0 || ink_rect->height == 0)
 	    {
-	      ink_rect->x = x_pos + glyph_ink.x;
+	      ink_rect->x = x_pos + glyph_ink.x + geometry->x_offset;
 	      ink_rect->width = glyph_ink.width;
-	      ink_rect->y = glyph_ink.y;
+	      ink_rect->y = glyph_ink.y + geometry->y_offset;
 	      ink_rect->height = glyph_ink.height;
 	    }
 	  else
@@ -263,6 +266,31 @@ pango_glyph_string_extents (PangoGlyphString *glyphs,
                                     font, ink_rect, logical_rect);
 }
 
+/**
+ * pango_glyph_string_get_width:
+ * @glyphs:   a #PangoGlyphString
+ * 
+ * Computes the logical width of the glyph string as can also be computed
+ * using pango_glyph_string_extents().  However, since this only computes the
+ * width, it's much faster.  This is in fact only a convenience function that
+ * computes the sum of geometry.width for each glyph in the @glyphs.
+ *
+ * Return value: the logical width of the glyph string.
+ *
+ * Since: 1.14
+ */
+int
+pango_glyph_string_get_width (PangoGlyphString *glyphs)
+{
+  int i;
+  int width = 0;
+
+  for (i = 0; i < glyphs->num_glyphs; i++)
+    width += glyphs->glyphs[i].geometry.width;
+
+  return width;
+}
+
 /**
  * pango_glyph_string_get_logical_widths:
  * @glyphs: a #PangoGlyphString
