@@ -19,11 +19,12 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <string.h>
 #include <stdlib.h>
 
-#include "pango/pango-context.h"
-#include "pango/pango-utils.h"
+#include "pango-context.h"
+#include "pango-impl-utils.h"
 
 #include "pango-engine.h"
 #include "pango-engine-private.h"
@@ -35,6 +36,9 @@ struct _PangoContext
 
   PangoLanguage *language;
   PangoDirection base_dir;
+  PangoGravity base_gravity;
+  PangoGravity resolved_gravity;
+
   PangoFontDescription *font_desc;
 
   PangoMatrix *matrix;
@@ -56,11 +60,12 @@ static void
 pango_context_init (PangoContext *context)
 {
   context->base_dir = PANGO_DIRECTION_WEAK_LTR;
+  context->resolved_gravity = context->base_gravity = PANGO_GRAVITY_SOUTH;
   context->language = NULL;
   context->font_map = NULL;
 
   context->font_desc = pango_font_description_new ();
-  pango_font_description_set_family (context->font_desc, "serif");
+  pango_font_description_set_family_static (context->font_desc, "serif");
   pango_font_description_set_style (context->font_desc, PANGO_STYLE_NORMAL);
   pango_font_description_set_variant (context->font_desc, PANGO_VARIANT_NORMAL);
   pango_font_description_set_weight (context->font_desc, PANGO_WEIGHT_NORMAL);
@@ -111,7 +116,8 @@ pango_context_finalize (GObject *object)
  * gdk_pango_context_get_for_screen(), and 
  * gtk_widget_get_pango_context().
  * 
- * Return value: the new #PangoContext
+ * Return value: the newly allocated #PangoContext, which should
+ *               be freed with g_object_unref().
  **/
 PangoContext *
 pango_context_new (void)
@@ -123,6 +129,15 @@ pango_context_new (void)
   return context;
 }
 
+static void
+update_resolved_gravity (PangoContext *context)
+{
+  if (context->base_gravity == PANGO_GRAVITY_AUTO)
+    context->resolved_gravity = pango_matrix_to_gravity (context->matrix);
+  else
+    context->resolved_gravity = context->base_gravity;
+}
+
 /**
  * pango_context_set_matrix:
  * @context: a #PangoContext
@@ -150,6 +165,8 @@ pango_context_set_matrix (PangoContext       *context,
     context->matrix = pango_matrix_copy (matrix);
   else
     context->matrix = NULL;
+
+  update_resolved_gravity (context);
 }
 
 /**
@@ -159,7 +176,7 @@ pango_context_set_matrix (PangoContext       *context,
  * Gets the transformation matrix that will be applied when
  * rendering with this context. See pango_context_set_matrix().
  *
- * Returns: the matrix, or %NULL if no matrix has been set
+ * Return value: the matrix, or %NULL if no matrix has been set
  *  (which is the same as the identity matrix). The returned
  *  matrix is owned by Pango and must not be modified or
  *  freed.
@@ -266,7 +283,6 @@ pango_context_load_font (PangoContext               *context,
 {
   g_return_val_if_fail (context != NULL, NULL);
   g_return_val_if_fail (pango_font_description_get_family (desc) != NULL, NULL);
-  g_return_val_if_fail (pango_font_description_get_size (desc) != 0, NULL);
   
   return pango_font_map_load_font (context->font_map, context, desc);
 }
@@ -334,7 +350,9 @@ pango_context_get_font_description (PangoContext *context)
  * @context: a #PangoContext
  * @language: the new language tag.
  * 
- * Sets the global language tag for the context.
+ * Sets the global language tag for the context.  The default language
+ * for the locale of the running process can be found using
+ * pango_language_get_default().
  **/
 void
 pango_context_set_language (PangoContext *context,
@@ -401,6 +419,68 @@ pango_context_get_base_dir (PangoContext *context)
   return context->base_dir;
 }
 
+/**
+ * pango_context_set_base_gravity:
+ * @context: a #PangoContext
+ * @gravity: the new base gravity
+ * 
+ * Sets the base gravity for the context.
+ *
+ * The base gravity is used in laying vertical text out.
+ *
+ * Since: 1.16
+ **/
+void
+pango_context_set_base_gravity (PangoContext  *context,
+				PangoGravity gravity)
+{
+  g_return_if_fail (context != NULL);
+
+  context->base_gravity = gravity;
+
+  update_resolved_gravity (context);
+}
+
+/**
+ * pango_context_get_base_gravity:
+ * @context: a #PangoContext
+ * 
+ * Retrieves the base gravity for the context. See
+ * pango_context_set_base_gravity().
+ * 
+ * Return value: the base gravity for the context.
+ *
+ * Since: 1.16
+ **/
+PangoGravity
+pango_context_get_base_gravity (PangoContext *context)
+{
+  g_return_val_if_fail (context != NULL, PANGO_GRAVITY_SOUTH);
+
+  return context->base_gravity;
+}
+
+/**
+ * pango_context_get_gravity:
+ * @context: a #PangoContext
+ * 
+ * Retrieves the gravity for the context. This is similar to
+ * pango_context_get_base_gravity(), except for when the base gravity
+ * is %PANGO_GRAVITY_AUTO for which pango_matrix_to_gravity() is used
+ * to return the gravity from the current context matrix.
+ * 
+ * Return value: the resolved gravity for the context.
+ *
+ * Since: 1.16
+ **/
+PangoGravity
+pango_context_get_gravity (PangoContext *context)
+{
+  g_return_val_if_fail (context != NULL, PANGO_GRAVITY_SOUTH);
+
+  return context->resolved_gravity;
+}
+
 /**********************************************************************/
 
 static gboolean
@@ -442,7 +522,7 @@ static void
 shaper_font_cache_destroy (ShaperFontCache *cache)
 {
   g_hash_table_destroy (cache->hash);
-  g_free (cache);
+  g_slice_free (ShaperFontCache, cache);
 }
 
 static void
@@ -450,7 +530,7 @@ shaper_font_element_destroy (ShaperFontElement *element)
 {
   if (element->font)
     g_object_unref (element->font);
-  g_free (element);
+  g_slice_free (ShaperFontElement, element);
 }
 
 static ShaperFontCache *
@@ -465,7 +545,7 @@ get_shaper_font_cache (PangoFontset *fontset)
   cache = g_object_get_qdata (G_OBJECT (fontset), cache_quark);
   if (!cache)
     {
-      cache = g_new (ShaperFontCache, 1);
+      cache = g_slice_new (ShaperFontCache);
       cache->hash = g_hash_table_new_full (g_direct_hash, NULL,
 					   NULL, (GDestroyNotify)shaper_font_element_destroy);
 
@@ -502,7 +582,7 @@ shaper_font_cache_insert (ShaperFontCache   *cache,
 			  PangoEngineShape  *shape_engine,
 			  PangoFont         *font)
 {
-  ShaperFontElement *element = g_new (ShaperFontElement, 1);
+  ShaperFontElement *element = g_slice_new (ShaperFontElement);
   element->shape_engine = shape_engine;
   element->font = font;
   
@@ -538,6 +618,10 @@ struct _ItemizeState
   const char *embedding_end;
   guint8 embedding;
 
+  PangoGravity gravity;
+  PangoGravity font_desc_gravity;
+  gboolean centered_baseline;
+
   PangoAttrIterator *attr_iter;
   gboolean free_attr_iter;
   const char *attr_end;
@@ -610,6 +694,11 @@ update_attr_iterator (ItemizeState *state)
   state->font_desc = pango_font_description_copy_static (state->context->font_desc);
   pango_attr_iterator_get_font (state->attr_iter, state->font_desc,
 				&state->lang, &state->extra_attrs);
+  if (pango_font_description_get_set_fields (state->font_desc) & PANGO_FONT_MASK_GRAVITY)
+    state->font_desc_gravity = pango_font_description_get_gravity (state->font_desc);
+  else
+    state->font_desc_gravity = PANGO_GRAVITY_AUTO;
+
   state->copy_extra_attrs = FALSE;
 
   if (!state->lang)
@@ -644,8 +733,6 @@ itemize_state_init (ItemizeState      *state,
 		    PangoAttrIterator *cached_iter,
 		    const PangoFontDescription *desc)
 {
-  gunichar *text_ucs4;
-  long n_chars;
 
   state->context = context;
   state->text = text;
@@ -664,7 +751,7 @@ itemize_state_init (ItemizeState      *state,
   state->embedding_end_offset = 0;
   state->embedding_end = text + start_index;
   update_embedding_end (state);
-  
+
   /* Initialize the attribute iterator
    */
   if (cached_iter)
@@ -710,6 +797,15 @@ itemize_state_init (ItemizeState      *state,
 
   update_end (state);
 
+  state->centered_baseline = state->context->resolved_gravity == PANGO_GRAVITY_EAST
+			  || state->context->resolved_gravity == PANGO_GRAVITY_WEST;
+
+  if (pango_font_description_get_set_fields (state->font_desc) & PANGO_FONT_MASK_GRAVITY)
+    state->font_desc_gravity = pango_font_description_get_gravity (state->font_desc);
+  else
+    state->font_desc_gravity = PANGO_GRAVITY_AUTO;
+
+  state->gravity = PANGO_GRAVITY_AUTO;
   state->derived_lang = NULL;
   state->lang_engine = NULL;
   state->current_fonts = NULL;
@@ -830,6 +926,41 @@ itemize_state_add_character (ItemizeState     *state,
   state->item->analysis.font = font;
   
   state->item->analysis.level = state->embedding;
+  state->item->analysis.gravity = state->gravity;
+
+  /* The level vs. gravity dance:
+   * 	- If gravity is SOUTH, leave level untouched.
+   * 	- If gravity is NORTH, step level one up, to
+   * 	  not get mirrored upside-down text.
+   * 	- If gravity is EAST, step up to an even level, as
+   * 	  it's a clockwise-rotated layout, so the rotated
+   * 	  top is unrotated left.
+   * 	- If gravity is WEST, step up to an odd level, as
+   * 	  it's a counter-clockwise-rotated layout, so the rotated
+   * 	  top is unrotated right.
+   *
+   * A similar dance is performed in pango-layout.c:
+   * line_set_resolved_dir().  Keep in synch.
+   */
+  switch (state->item->analysis.gravity)
+    {
+      case PANGO_GRAVITY_SOUTH:
+      default:
+	break;
+      case PANGO_GRAVITY_NORTH:
+	state->item->analysis.level++;
+	break;
+      case PANGO_GRAVITY_EAST:
+	state->item->analysis.level += 1;
+	state->item->analysis.level &= ~1;
+	break;
+      case PANGO_GRAVITY_WEST:
+	state->item->analysis.level |= 1;
+	break;
+    }
+
+  state->item->analysis.centered_baseline = state->centered_baseline;
+
   state->item->analysis.language = state->derived_lang;
   
   if (state->copy_extra_attrs)
@@ -918,8 +1049,29 @@ get_shaper_and_font (ItemizeState      *state,
     return *shape_engine != NULL;
 
   if (!state->exact_engines && !state->fallback_engines)
-    get_engines (state->context, state->derived_lang, state->script,
-		 &state->exact_engines, &state->fallback_engines);
+    {
+      /* Always use a basic shaper for vertical layout (ie, east/west gravity)
+       * as we do not support vertical shaping as of now.
+       */
+      PangoScript script;
+
+      switch (state->gravity)
+	{
+	  case PANGO_GRAVITY_SOUTH:
+	  case PANGO_GRAVITY_NORTH:
+	  case PANGO_GRAVITY_AUTO:
+	  default:
+            script = state->script;
+	    break;
+	  case PANGO_GRAVITY_EAST:
+	  case PANGO_GRAVITY_WEST:
+            script = PANGO_SCRIPT_COMMON;
+	    break;
+	}
+
+      get_engines (state->context, state->derived_lang, script,
+		   &state->exact_engines, &state->fallback_engines);
+    }
 
   info.lang = state->derived_lang;
   info.wc = wc;
@@ -1016,6 +1168,30 @@ get_lang_map (PangoLanguage *lang)
 static void
 itemize_state_update_for_new_run (ItemizeState *state)
 {
+  if (state->changed & (FONT_CHANGED | SCRIPT_CHANGED))
+    {
+      PangoGravity old_gravity = state->gravity;
+
+      if (state->font_desc_gravity != PANGO_GRAVITY_AUTO)
+	state->gravity = state->font_desc_gravity;
+      else
+        {
+	  PangoGravity gravity = state->context->resolved_gravity;
+
+	  /* FIXME
+	   * gravity = pango_script_get_resolved_gravity (script, gravity, hint);
+	   */
+
+	  state->gravity = gravity;
+	}
+
+      if (old_gravity != state->gravity)
+        {
+	  pango_font_description_set_gravity (state->font_desc, state->gravity);
+          state->changed |= FONT_CHANGED;
+	}
+    }
+
   if (state->changed & (SCRIPT_CHANGED | LANG_CHANGED))
     {
       PangoLanguage *old_derived_lang = state->derived_lang;
@@ -1082,24 +1258,36 @@ itemize_state_process_run (ItemizeState *state)
       PangoFont *font;
 
       /* We don't want space characters to affect font selection; in general,
-       * it's always wrong to select a font just to render a space.
-       * To make this really work well, we'd need to emulate missing
-       * space characters, which we don't yet do, so we sometimes
-       * will get hex boxes.
+       * it's always wrong to select a font just to render a space.  But until
+       * we have a better solution, choosing a font for spaces seems to work
+       * better.
        *
        * The exception of U+3000 (IDEOGRAPHIC SPACE) here is because we
        * want to choose an ideographic space that matches ideographic text
        * in cell width. Even if we were emulating missing spaces, an
        * emulated ideographic space for the primary font wouldn't be the
        * right size.
+       *
+       * The exception of PrivateUse and Unassigned characters is necessary
+       * to be able to render any of them. (for private or being encoded
+       * scripts, etc.)
        */
-      if (!g_unichar_isgraph (wc) && wc != 0x3000)
+      /*
+      if (G_UNLIKELY (!g_unichar_isgraph (wc) && wc != 0x3000 &&
+		      g_unichar_type (wc) != G_UNICODE_PRIVATE_USE &&
+		      g_unichar_type (wc) != G_UNICODE_UNASSIGNED))
+	*/
+      if (G_UNLIKELY (!g_unichar_isprint (wc) &&
+		      g_unichar_type (wc) != G_UNICODE_PRIVATE_USE &&
+		      g_unichar_type (wc) != G_UNICODE_UNASSIGNED))
 	{
 	  shape_engine = NULL;
 	  font = NULL;
 	}
       else
-        get_shaper_and_font (state, wc, &shape_engine, &font);
+        {
+	  get_shaper_and_font (state, wc, &shape_engine, &font);
+	}
 	
       itemize_state_add_character (state,
 				   shape_engine, font,
@@ -1310,7 +1498,7 @@ update_metrics_from_items (PangoFontMetrics *metrics,
       PangoItem *item = l->data;
       PangoFont *font = item->analysis.font;
 
-      if (g_hash_table_lookup (fonts_seen, font) == NULL)
+      if (font != NULL && g_hash_table_lookup (fonts_seen, font) == NULL)
 	{
 	  PangoFontMetrics *raw_metrics = pango_font_get_metrics (font, language);
 	  g_hash_table_insert (fonts_seen, font, font);
@@ -1367,7 +1555,7 @@ update_metrics_from_items (PangoFontMetrics *metrics,
  * be a composite of the metrics for the fonts loaded for the
  * individual families.
  *
- * Returns: a #PangoFontMetrics object. The caller must call pango_font_metrics_unref()
+ * Return value: a #PangoFontMetrics object. The caller must call pango_font_metrics_unref()
  *   when finished using the object.
  **/
 PangoFontMetrics *
