@@ -24,8 +24,10 @@
 #include <string.h>
 #include <stdlib.h>
 #include <math.h>
+#include <locale.h>
 
 #include "pango-font.h"
+#include "pango-features.h"
 #include "pango-impl-utils.h"
 
 #include <glib/gstdio.h>
@@ -59,6 +61,93 @@ static GHashTable *pango_aliases_ht = NULL;
 
 PangoWarningHistory _pango_warning_history = { FALSE, FALSE, FALSE };
 
+/**
+ * pango_version:
+ *
+ * This is similar to the macro %PANGO_VERSION except that
+ * it returns the encoded version of Pango available at run-time,
+ * as opposed to the version available at compile-time.
+ *
+ * A version number can be encoded into an integer using
+ * PANGO_VERSION_ENCODE().
+ *
+ * Returns value: The encoded version of Pango library
+ *   available at run time.
+ *
+ * Since: 1.16
+ **/
+int
+pango_version (void)
+{
+  return PANGO_VERSION;
+}
+
+/**
+ * pango_version_string:
+ *
+ * This is similar to the macro %PANGO_VERSION_STRING except that
+ * it returns the version of Pango available at run-time, as opposed to
+ * the version available at compile-time.
+ *
+ * Returns value: A string containing the version of Pango library
+ *   available at run time.
+ *   The returned string is owned by Pango and should not be modified
+ *   or freed.
+ *
+ * Since: 1.16
+ **/
+const char *
+pango_version_string (void)
+{
+  return PANGO_VERSION_STRING;
+}
+
+/**
+ * pango_version_check:
+ * @required_major: the required major version.
+ * @required_minor: the required minor version.
+ * @required_micro: the required major version.
+ *
+ * Checks that the Pango library in use is compatible with the
+ * given version. Generally you would pass in the constants
+ * %PANGO_VERSION_MAJOR, %PANGO_VERSION_MINOR, %PANGO_VERSION_MICRO
+ * as the three arguments to this function; that produces
+ * a check that the library in use at run-time is compatible with
+ * the version of Pango the application or module was compiled against.
+ *
+ * Compatibility is defined by two things: first the version
+ * of the running library is newer than the version
+ * @required_major.required_minor.@required_micro. Second
+ * the running library must be binary compatible with the
+ * version @required_major.required_minor.@required_micro
+ * (same major version.)
+ *
+ * For compile-time version checking use PANGO_VERSION_CHECK().
+ *
+ * Return value: %NULL if the Pango library is compatible with the
+ *   given version, or a string describing the version mismatch.
+ *   The returned string is owned by Pango and should not be modified
+ *   or freed.
+ *
+ * Since: 1.16
+ **/
+const gchar*
+pango_version_check (int required_major,
+                     int required_minor,
+                     int required_micro)
+{
+  gint pango_effective_micro = 100 * PANGO_VERSION_MINOR + PANGO_VERSION_MICRO;
+  gint required_effective_micro = 100 * required_minor + required_micro;
+
+  if (required_major < PANGO_VERSION_MAJOR)
+    return "Pango version too new (major mismatch)";
+  if (required_effective_micro < pango_effective_micro - PANGO_BINARY_AGE)
+    return "Pango version too new (micro mismatch)";
+  if (required_effective_micro > pango_effective_micro)
+    return "Pango version too old (micro mismatch)";
+  return NULL;
+}
+
 /**
  * pango_trim_string:
  * @str: a string
@@ -987,6 +1076,95 @@ pango_language_get_type (void)
   return our_type;
 }
 
+/**
+ * _pango_get_lc_ctype:
+ *
+ * Return the Unix-style locale string for the language currently in
+ * effect. On Unix systems, this is the return value from
+ * <literal>setlocale(LC_CTYPE, NULL)</literal>, and the user can
+ * affect this through the environment variables LC_ALL, LC_CTYPE or
+ * LANG (checked in that order). The locale strings typically is in
+ * the form lang_COUNTRY, where lang is an ISO-639 language code, and
+ * COUNTRY is an ISO-3166 country code. For instance, sv_FI for
+ * Swedish as written in Finland or pt_BR for Portuguese as written in
+ * Brazil.
+ * 
+ * On Windows, the C library doesn't use any such environment
+ * variables, and setting them won't affect the behaviour of functions
+ * like ctime(). The user sets the locale through the Regional Options 
+ * in the Control Panel. The C library (in the setlocale() function) 
+ * does not use country and language codes, but country and language 
+ * names spelled out in English. 
+ * However, this function does check the above environment
+ * variables, and does return a Unix-style locale string based on
+ * either said environment variables or the thread's current locale.
+ *
+ * Return value: a dynamically allocated string, free with g_free().
+ */
+static gchar *
+_pango_get_lc_ctype (void)
+{
+#ifdef G_OS_WIN32
+  /* Somebody might try to set the locale for this process using the
+   * LANG or LC_ environment variables. The Microsoft C library
+   * doesn't know anything about them. You set the locale in the
+   * Control Panel. Setting these env vars won't have any affect on
+   * locale-dependent C library functions like ctime(). But just for
+   * kicks, do obey LC_ALL, LC_CTYPE and LANG in Pango. (This also makes
+   * it easier to test GTK and Pango in various default languages, you
+   * don't have to clickety-click in the Control Panel, you can simply
+   * start the program with LC_ALL=something on the command line.)
+   */
+
+  gchar *p;
+
+  p = getenv ("LC_ALL");
+  if (p != NULL)
+    return g_strdup (p);
+
+  p = getenv ("LC_CTYPE");
+  if (p != NULL)
+    return g_strdup (p);
+
+  p = getenv ("LANG");
+  if (p != NULL)
+    return g_strdup (p);
+
+  return g_win32_getlocale ();
+#else
+  return g_strdup (setlocale (LC_CTYPE, NULL));
+#endif
+}
+
+/**
+ * pango_language_get_default:
+ *
+ * Returns the #PangoLanguage for the current locale of the process.
+ * Note that this can change over the life of an application.
+ *
+ * Your application should call <literal>setlocale (LC_ALL, "");</literal>
+ * for the user settings to take effect.  Gtk+ does this in its initialization
+ * functions.  See <literal>man setlocale</literal> for details.
+ * 
+ * Return value: the default language as a #PangoLanguage, must not be
+ *               freed.
+ *
+ * Since: 1.16
+ **/
+PangoLanguage *
+pango_language_get_default (void)
+{
+  gchar *lang;
+  PangoLanguage *result;  
+  
+  lang = _pango_get_lc_ctype ();
+ 
+  result = pango_language_from_string (lang);
+  g_free (lang);
+  
+  return result;
+}
+
 /**
  * pango_language_from_string:
  * @language: a string representing a language tag
@@ -999,6 +1177,9 @@ pango_language_get_type (void)
  * This function first canonicalizes the string by converting it to
  * lowercase, mapping '_' to '-', and stripping all characters other
  * than letters and '-'.
+ *
+ * Use pango_language_get_default() if you want to get the #PangoLanguage for
+ * the current locale of the process.
  * 
  * Return value: an opaque pointer to a #PangoLanguage structure.
  *               this will be valid forever after.
@@ -1634,7 +1815,7 @@ pango_quantize_line_geometry (int *thickness,
  * Converts a #PangoGravity value to its rotation value.
  *
  * Return value: the rotation value corresponding to @gravity,
- * or zero if @gravity is #PANGO_GRAVITY_AUTO
+ * or zero if @gravity is %PANGO_GRAVITY_AUTO
  *
  * Since: 1.16
  */
@@ -1654,3 +1835,4 @@ pango_gravity_to_rotation (PangoGravity gravity)
 
   return rotation;
 }
+
