@@ -19,12 +19,13 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <string.h>
 #include <stdlib.h>
 #include <math.h>
 
 #include <X11/Xlib.h>
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 #include "modules.h"
 
 #undef PANGO_DISABLE_DEPRECATED
@@ -32,10 +33,6 @@
 #include "pangox.h"
 #include "pangox-private.h"
 
-#define PANGO_X_UNKNOWN_FLAG 0x10000000
-
-#include "config.h"
-
 #define PANGO_TYPE_X_FONT              (pango_x_font_get_type ())
 #define PANGO_X_FONT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_X_FONT, PangoXFont))
 #define PANGO_X_FONT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_X_FONT, PangoXFontClass))
@@ -92,6 +89,7 @@ static void                  pango_x_font_get_glyph_extents (PangoFont        *f
 							     PangoRectangle   *logical_rect);
 static PangoFontMetrics *    pango_x_font_get_metrics       (PangoFont        *font,
 							     PangoLanguage    *language);
+static PangoFontMap *        pango_x_font_get_font_map      (PangoFont        *font);
 
 static PangoXSubfontInfo * pango_x_find_subfont    (PangoFont          *font,
 						    PangoXSubfont       subfont_index);
@@ -176,7 +174,7 @@ pango_x_get_context (Display *display)
   static gboolean registered_modules = FALSE;
   
   g_return_val_if_fail (display != NULL, NULL);  
-  
+
   if (!registered_modules)
     {
       registered_modules = TRUE;
@@ -231,7 +229,7 @@ pango_x_font_get_type (void)
 
   if (!object_type)
     {
-      static const GTypeInfo object_info =
+      const GTypeInfo object_info =
       {
         sizeof (PangoXFontClass),
         (GBaseInitFunc) NULL,
@@ -242,10 +240,11 @@ pango_x_font_get_type (void)
         sizeof (PangoXFont),
         0,              /* n_preallocs */
         (GInstanceInitFunc) pango_x_font_init,
+	NULL            /* value_table */
       };
       
       object_type = g_type_register_static (PANGO_TYPE_FONT,
-                                            "PangoXFont",
+                                            I_("PangoXFont"),
                                             &object_info, 0);
     }
   
@@ -283,6 +282,7 @@ pango_x_font_class_init (PangoXFontClass *class)
   font_class->find_shaper = pango_x_font_find_shaper;
   font_class->get_glyph_extents = pango_x_font_get_glyph_extents;
   font_class->get_metrics = pango_x_font_get_metrics;
+  font_class->get_font_map = pango_x_font_get_font_map;
 }
 
 PangoXFont *
@@ -399,31 +399,85 @@ pango_x_render  (Display           *display,
        * the ink rect here would be a noticeable speed hit.
        * This is close enough.
        */
-      if (!(glyph &&
+      if (!(glyph != PANGO_GLYPH_EMPTY &&
 	    glyph_x >= -16384 && glyph_x <= 32767 &&
 	    glyph_y >= -16384 && glyph_y <= 32767))
 	goto next_glyph;
 	      
-      if (glyph & PANGO_X_UNKNOWN_FLAG)
-        {
-	  PangoFontMetrics *metrics = pango_font_get_metrics (font,
-							      pango_language_from_string ("en"));
+      if ((glyph & PANGO_GLYPH_UNKNOWN_FLAG) == 0)
+	{
+	  guint16 index = PANGO_X_GLYPH_INDEX (glyph);
+	  guint16 subfont_index = PANGO_X_GLYPH_SUBFONT (glyph);
+	  PangoXSubfontInfo *subfont;
+      
+	  subfont = pango_x_find_subfont (font, subfont_index);
+	  if (subfont)
+	    {
+	      fs = pango_x_get_font_struct (font, subfont);
+	      if (!fs)
+		continue;
+	      
+	      if (fs->fid != old_fid)
+		{
+		  FLUSH;
+		  XSetFont (display, gc, fs->fid);
+		  old_fid = fs->fid;
+		}
+
+	      if (charcount == G_N_ELEMENTS (xcharbuffer) ||
+		  (charcount > 0 && (glyph_y != glyph_y0 ||
+				     glyph_x != expected_x)))
+		FLUSH;
+
+	      if (charcount == 0)
+		{
+		  glyph_x0 = glyph_x;
+		  glyph_y0 = glyph_y;
+		}
+	      xcharbuffer[charcount].byte1 = index / 256;
+	      xcharbuffer[charcount].byte2 = index % 256;
+
+	      expected_x = glyph_x + XTextWidth16 (fs, &xcharbuffer[charcount], 1);
+
+	      charcount++;
+	    } else
+	      goto unknown_glyph;
+	} else {
+	  PangoFontMetrics *metrics;
           int x1, y1, x2, y2; /* rectangle the character should go inside. */
           int baseline;
 	  int stroke_thick;
-
           gunichar wc;
-
+	  
+	unknown_glyph:
 	  FLUSH;
 
-          x1 = glyph_x;
-          y1 = glyph_y - PANGO_PIXELS (metrics->ascent);
-          x2 = x1 + PANGO_PIXELS (glyphs->glyphs[i].geometry.width);
-          y2 = y1 + PANGO_PIXELS (metrics->ascent + metrics->descent);
+	  if (font)
+	    metrics = pango_font_get_metrics (font, NULL);
+	  else
+	    metrics = NULL;
+
+	  if (metrics)
+	    {
+	      y1 = glyph_y - PANGO_PIXELS (metrics->ascent);
+	      y2 = y1 + PANGO_PIXELS (metrics->ascent + metrics->descent);
+	    }
+	  else
+	    {
+	      y2 = glyph_y;
+	      y1 = y2 - PANGO_UNKNOWN_GLYPH_HEIGHT;
+	    }
+
+	  x1 = glyph_x;
+	  x2 = x1 + PANGO_PIXELS (glyphs->glyphs[i].geometry.width);
+
           baseline = glyph_y;
 	  stroke_thick = MAX ((int) (0.5 + 0.075 * (y2 - y1)), 1);
           
-          wc = glyph & (~PANGO_X_UNKNOWN_FLAG);
+	  if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
+	    wc = glyph & (~PANGO_GLYPH_UNKNOWN_FLAG);
+	  else
+	    wc = 0;
 
           switch (wc)
             {
@@ -518,44 +572,6 @@ pango_x_render  (Display           *display,
 
 	  pango_font_metrics_unref (metrics);
         }
-      else
-	{
-	  guint16 index = PANGO_X_GLYPH_INDEX (glyph);
-	  guint16 subfont_index = PANGO_X_GLYPH_SUBFONT (glyph);
-	  PangoXSubfontInfo *subfont;
-      
-	  subfont = pango_x_find_subfont (font, subfont_index);
-	  if (subfont)
-	    {
-	      fs = pango_x_get_font_struct (font, subfont);
-	      if (!fs)
-		continue;
-	      
-	      if (fs->fid != old_fid)
-		{
-		  FLUSH;
-		  XSetFont (display, gc, fs->fid);
-		  old_fid = fs->fid;
-		}
-
-	      if (charcount == G_N_ELEMENTS (xcharbuffer) ||
-		  (charcount > 0 && (glyph_y != glyph_y0 ||
-				     glyph_x != expected_x)))
-		FLUSH;
-
-	      if (charcount == 0)
-		{
-		  glyph_x0 = glyph_x;
-		  glyph_y0 = glyph_y;
-		}
-	      xcharbuffer[charcount].byte1 = index / 256;
-	      xcharbuffer[charcount].byte2 = index % 256;
-
-	      expected_x = glyph_x + XTextWidth16 (fs, &xcharbuffer[charcount], 1);
-
-	      charcount++;
-	    }
-	}
 
     next_glyph:
       x_off += glyphs->glyphs[i].geometry.width;
@@ -574,56 +590,15 @@ pango_x_font_get_glyph_extents  (PangoFont      *font,
   XCharStruct *cs;
   PangoXSubfontInfo *subfont;
 
-  if (glyph & PANGO_X_UNKNOWN_FLAG)
+  if (glyph == PANGO_GLYPH_EMPTY)
     {
-      PangoFontMetrics *metrics = pango_font_get_metrics (font,
-							  pango_language_from_string ("en"));
-      gunichar wc;
-      gdouble width_factor;
-      int w;
-      
-      wc = glyph & (~PANGO_X_UNKNOWN_FLAG);
-          
-      switch (wc)
-        {
-        case '\n':
-        case '\r':
-        case 0x2028: /* Line separator */
-        case 0x2029: /* Paragraph separator */
-          {
-#define MAGIC_FACTOR 1.2
-
-            /* carriage-return thingy */
-	    width_factor = MAGIC_FACTOR;
-	    break;
-	  }
-	default:
-	  {
-	    /* Unknown glyph square */
-	    width_factor = 0.7;
-	  }
-	}
-	    
-      w = metrics->approximate_char_width * width_factor;
-      w = PANGO_SCALE * PANGO_PIXELS (w);
-            
       if (ink_rect)
-	{
-	  ink_rect->x = 0;
-	  ink_rect->width = w;
-	  ink_rect->y = - metrics->ascent;
-	  ink_rect->height = metrics->ascent + metrics->descent;
-	}
+	ink_rect->x = ink_rect->width = ink_rect->y = ink_rect->height = 0;
       if (logical_rect)
-	{
-	  logical_rect->x = 0;
-	  logical_rect->width = w;
-	  logical_rect->y = - metrics->ascent;
-	  logical_rect->height = metrics->ascent + metrics->descent;
-	}
-      
+	logical_rect->x = logical_rect->width = logical_rect->y = logical_rect->height = 0;
+      return;
     }
-  else if (glyph && pango_x_find_glyph (font, glyph, &subfont, &cs))
+  if ((glyph & PANGO_GLYPH_UNKNOWN_FLAG) == 0 && pango_x_find_glyph (font, glyph, &subfont, &cs))
     {
       if (ink_rect)
 	{
@@ -642,19 +617,66 @@ pango_x_font_get_glyph_extents  (PangoFont      *font,
     }
   else
     {
-      if (ink_rect)
-	{
-	  ink_rect->x = 0;
-	  ink_rect->width = 0;
-	  ink_rect->y = 0;
-	  ink_rect->height = 0;
+      PangoFontMetrics *metrics;
+      gunichar wc;
+      gdouble width_factor;
+      int w;
+      
+      if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
+	wc = glyph & (~PANGO_GLYPH_UNKNOWN_FLAG);
+      else
+        wc = 0;
+          
+      switch (wc)
+        {
+        case '\n':
+        case '\r':
+        case 0x2028: /* Line separator */
+        case 0x2029: /* Paragraph separator */
+          {
+#define MAGIC_FACTOR 1.2
+
+            /* carriage-return thingy */
+	    width_factor = MAGIC_FACTOR;
+	    break;
+	  }
+	default:
+	  {
+	    /* Unknown glyph square */
+	    width_factor = 1.0;
+	  }
 	}
-      if (logical_rect)
+	    
+      metrics = pango_font_get_metrics (font, NULL);
+
+      if (metrics)
+        {
+	  w = metrics->approximate_char_width * width_factor;
+	  w = PANGO_SCALE * PANGO_PIXELS (w);
+		
+	  if (ink_rect)
+	    {
+	      ink_rect->x = PANGO_SCALE;
+	      ink_rect->width = w - 2 * PANGO_SCALE;
+	      ink_rect->y = - (metrics->ascent - PANGO_SCALE);
+	      ink_rect->height = metrics->ascent + metrics->descent - 2 * PANGO_SCALE;
+	    }
+	  if (logical_rect)
+	    {
+	      logical_rect->x = 0;
+	      logical_rect->width = w;
+	      logical_rect->y = - metrics->ascent;
+	      logical_rect->height = metrics->ascent + metrics->descent;
+	    }
+
+	  pango_font_metrics_unref (metrics);
+	}
+      else
 	{
-	  logical_rect->x = 0;
-	  logical_rect->width = 0;
-	  logical_rect->y = 0;
-	  logical_rect->height = 0;
+	  if (ink_rect)
+	    ink_rect->x = ink_rect->y = ink_rect->height = ink_rect->width = 0;
+	  if (logical_rect)
+	    logical_rect->x = logical_rect->y = logical_rect->height = logical_rect->width = 0;
 	}
     }
 }
@@ -701,20 +723,12 @@ itemize_string_foreach (PangoFont     *font,
   PangoGlyphString *glyph_str = pango_glyph_string_new ();
   PangoEngineShape *shaper, *last_shaper;
   int last_level;
-  gunichar *text_ucs4;
-  long n_chars, i;
+  int i;
   guint8 *embedding_levels;
   PangoDirection base_dir = PANGO_DIRECTION_LTR;
   gboolean finished = FALSE;
   
-  text_ucs4 = g_utf8_to_ucs4_fast (str, -1, &n_chars);
-  if (!text_ucs4)
-    return;
-
-  embedding_levels = g_new (guint8, n_chars);
-  pango_log2vis_get_embedding_levels (text_ucs4, n_chars, &base_dir,
-				      embedding_levels);
-  g_free (text_ucs4);
+  embedding_levels = pango_log2vis_get_embedding_levels (str, -1, &base_dir);
 
   last_shaper = NULL;
   last_level = 0;
@@ -799,7 +813,7 @@ get_font_metrics_from_subfonts (PangoFont        *font,
       if (subfont)
 	{
 	  XFontStruct *fs = pango_x_get_font_struct (font, subfont);
-          gint avg_width;
+          gint avg_width = 0;
           
 	  if (fs)
 	    {
@@ -814,28 +828,31 @@ get_font_metrics_from_subfonts (PangoFont        *font,
 		  metrics->ascent = MAX (fs->ascent * PANGO_SCALE, metrics->ascent);
 		  metrics->descent = MAX (fs->descent * PANGO_SCALE, metrics->descent);
 		}
+
+	      if (get_int_prop (avg_width_atom, fs, &avg_width))
+		{
+		  /* convert decipoints --> Pango units.
+		   * Resolution is in (points * PANGO_SCALE) / pixel,
+		   * avg_width in decipoints.
+		   * We want pixels * PANGO_SCALE
+		   */
+
+		  /* Convert to points * PANGO_SCALE */
+		  avg_width *= PANGO_SCALE / (double) 10.0;
+		  /* Convert to pixels * PANGO_SCALE */
+		  avg_width *= (PANGO_SCALE / PANGO_X_FONT_MAP (PANGO_X_FONT (font)->fontmap)->resolution);
+		}
+	      else
+		{
+		  avg_width = PANGO_SCALE * ((fs->min_bounds.width + fs->max_bounds.width) / 2);
+		}
 	    }
-
-          if (get_int_prop (avg_width_atom, fs, &avg_width))
-            {
-              /* convert decipoints --> pango units.
-               * Resolution is in (points * PANGO_SCALE) / pixel,
-               * avg_width in decipoints.
-               * We want pixels * PANGO_SCALE
-               */
-
-              /* Convert to points * PANGO_SCALE */
-              avg_width *= PANGO_SCALE / (double) 10.0;
-              /* Convert to pixels * PANGO_SCALE */
-              avg_width *= (PANGO_SCALE / PANGO_X_FONT_MAP (PANGO_X_FONT (font)->fontmap)->resolution);
-            }
-          else
-            {
-              avg_width = PANGO_SCALE * ((fs->min_bounds.width + fs->max_bounds.width) / 2);
-            }
           
-          total_avg_widths += avg_width;
-          n_avg_widths += 1;
+	  if (avg_width)
+	    {
+	      total_avg_widths += avg_width;
+	      n_avg_widths += 1;
+	    }
 	}
       else
 	g_warning ("Invalid subfont %d in get_font_metrics_from_subfonts", GPOINTER_TO_UINT (tmp_list->data));
@@ -847,7 +864,12 @@ get_font_metrics_from_subfonts (PangoFont        *font,
   if (n_avg_widths)
     metrics->approximate_char_width = total_avg_widths / n_avg_widths;
   else
-    metrics->approximate_char_width = 10 * PANGO_SCALE;
+    metrics->approximate_char_width = PANGO_UNKNOWN_GLYPH_WIDTH * PANGO_SCALE;
+  if (metrics->ascent + metrics->descent == 0)
+    {
+      metrics->ascent = PANGO_UNKNOWN_GLYPH_HEIGHT * PANGO_SCALE;
+      metrics->descent = 0;
+    }
 }
 
 static void
@@ -857,13 +879,23 @@ get_subfonts_foreach (PangoFont      *font,
 {
   GSList **subfonts = data;
   PangoGlyph glyph = glyph_info->glyph;
+  PangoXSubfont subfont;
 
-  if ((glyph & PANGO_X_UNKNOWN_FLAG) == 0)
+  if (glyph == PANGO_GLYPH_EMPTY)
+    return;
+
+  /* Use an arbitrary subfont for unknown glyphs...*/
+  if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
     {
-      PangoXSubfont subfont = PANGO_X_GLYPH_SUBFONT (glyph);
-      if (!g_slist_find (*subfonts, GUINT_TO_POINTER ((guint)subfont)))
-	*subfonts = g_slist_prepend (*subfonts, GUINT_TO_POINTER ((guint)subfont));
+    if (((PangoXFont *)font)->n_subfonts > 0)
+      glyph = PANGO_X_MAKE_GLYPH (1, 0);
+    else
+      return;
     }
+
+  subfont = PANGO_X_GLYPH_SUBFONT (glyph);
+  if (!g_slist_find (*subfonts, GUINT_TO_POINTER ((guint)subfont)))
+    *subfonts = g_slist_prepend (*subfonts, GUINT_TO_POINTER ((guint)subfont));
 }
 
 /* Get composite font metrics for all subfonts resulting from shaping
@@ -930,23 +962,36 @@ pango_x_font_get_metrics (PangoFont        *font,
 
   if (!tmp_list)
     {
-      info = g_new (PangoXMetricsInfo, 1);
-      info->sample_str = sample_str;
-      info->metrics = pango_font_metrics_new ();
+      PangoFontMetrics *metrics;
+
+      info = g_new0 (PangoXMetricsInfo, 1);
 
       xfont->metrics_by_lang = g_slist_prepend (xfont->metrics_by_lang, info);
 
-      get_font_metrics_from_string (font, language, sample_str, info->metrics);
+      info->sample_str = sample_str;
+      metrics = pango_font_metrics_new ();
 
-      info->metrics->approximate_digit_width = get_total_width_for_string (font, language, "0123456789") / 10;
+      get_font_metrics_from_string (font, language, sample_str, metrics);
+
+      metrics->approximate_digit_width = get_total_width_for_string (font, language, "0123456789") / 10;
+
+      info->metrics = metrics;
     }
       
   return pango_font_metrics_ref (info->metrics);
 }
 
+static PangoFontMap *
+pango_x_font_get_font_map (PangoFont *font)
+{
+  PangoXFont *xfont = (PangoXFont *)font;
+
+  return xfont->fontmap;
+}
+
 /* Compare the tail of a to b */
 static gboolean
-match_end (char *a, char *b)
+match_end (const char *a, const char *b)
 {
   size_t len_a = strlen (a);
   size_t len_b = strlen (b);
@@ -958,7 +1003,7 @@ match_end (char *a, char *b)
 }
 
 /* Substitute in a charset into an XLFD. Return the
- * (g_malloc'd) new name, or NULL if the XLFD cannot
+ * (g_malloc'd) new name, or %NULL if the XLFD cannot
  * match the charset
  */
 static char *
@@ -1316,9 +1361,11 @@ pango_x_font_find_shaper (PangoFont     *font,
 			  guint32        ch)
 {
   PangoMap *shape_map = NULL;
+  PangoScript script;
 
   shape_map = pango_x_get_shaper_map (language);
-  return (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
+  script = pango_script_for_unichar (ch);
+  return (PangoEngineShape *)pango_map_get_engine (shape_map, script);
 }
 
 /* Utility functions */
@@ -1400,7 +1447,7 @@ pango_x_find_glyph (PangoFont *font,
  * @font: a #PangoFont.
  * 
  * Returns the index of a glyph suitable for drawing unknown characters;
- * you should generally use pango_x_font_get_unknown_glyph_instead,
+ * you should generally use PANGO_GET_UNKNOWN_GLYPH() instead,
  * since that may return a glyph that provides a better representation
  * of a particular char. (E.g., by showing hex digits, or a glyph
  * representive of a certain Unicode range.)
@@ -1410,7 +1457,7 @@ pango_x_find_glyph (PangoFont *font,
 PangoGlyph
 pango_x_get_unknown_glyph (PangoFont *font)
 {
-  return PANGO_X_UNKNOWN_FLAG;
+  return PANGO_GET_UNKNOWN_GLYPH (0);
 }
 
 /**
@@ -1544,7 +1591,7 @@ pango_x_render_layout_line (Display          *display,
  * @x:         the x position of the left of the layout (in pixels).
  * @y:         the y position of the top of the layout (in pixels).
  *
- * Renders a #PangoLayoutLine onto an X drawable.
+ * Renders a #PangoLayout onto an X drawable.
  */
 void 
 pango_x_render_layout (Display         *display,
@@ -1707,7 +1754,7 @@ pango_x_find_first_subfont (PangoFont      *font,
  * 
  * This is a simple fallback shaper, that can be used
  * if no subfont that supports a given script is found. 
- * For every character in @text, it puts the Unknown glyph.
+ * For every character in @text, it puts the unknown glyph.
  */
 void 
 pango_x_fallback_shape (PangoFont        *font, 
@@ -1731,10 +1778,12 @@ pango_x_fallback_shape (PangoFont        *font,
   for (i = 0; i < n_chars; i++)
     {
       glyphs->glyphs[i].glyph = unknown_glyph;
+
       glyphs->glyphs[i].geometry.x_offset = 0;
       glyphs->glyphs[i].geometry.y_offset = 0;
       glyphs->glyphs[i].geometry.width = logical_rect.width;
-      glyphs->log_clusters[i] = 0;
+
+      glyphs->log_clusters[i] = p - text;
       
       p = g_utf8_next_char (p);
     }
@@ -1748,13 +1797,13 @@ pango_x_fallback_shape (PangoFont        *font,
  * Returns the index of a glyph suitable for drawing @wc as an
  * unknown character.
  *
+ * Use PANGO_GET_UNKNOWN_GLYPH() instead.
+ *
  * Return value: a glyph index into @font.
  */
 PangoGlyph
 pango_x_font_get_unknown_glyph (PangoFont *font,
                                 gunichar   wc)
 {
-  g_return_val_if_fail (PANGO_IS_FONT (font), 0);
-
-  return PANGO_X_UNKNOWN_FLAG | wc;
+  return PANGO_GET_UNKNOWN_GLYPH (wc);
 }
